@@ -19,9 +19,11 @@
 
 #include "indicator-ng.h"
 #include "indicator-image-helper.h"
-
+#include <libayatana-ido/ayatanamenuitemfactory.h>
 #include <string.h>
 
+#define MENU_SECTIONS 20
+
 struct _IndicatorNg
 {
   IndicatorObject parent;
@@ -48,6 +50,7 @@ struct _IndicatorNg
   gchar *accessible_desc;
 
   gint64 last_service_restart;
+    GMenuModel *lMenuSections[MENU_SECTIONS];
 };
 
 static void indicator_ng_initable_iface_init (GInitableIface *initable);
@@ -62,6 +65,7 @@ enum
   N_PROPERTIES
 };
 
+static GQuark m_pActionMuxer = 0;
 static GParamSpec *properties[N_PROPERTIES];
 
 static void
@@ -122,6 +126,15 @@ indicator_ng_free_actions_and_menu (IndicatorNg *self)
 
   if (self->menu)
     {
+        for (guint nMenuSection = 0; nMenuSection < MENU_SECTIONS; nMenuSection++)
+        {
+            if (self->lMenuSections[nMenuSection])
+            {
+                g_object_unref(self->lMenuSections[nMenuSection]);
+                self->lMenuSections[nMenuSection] = NULL;
+            }
+        }
+
       g_signal_handlers_disconnect_by_data (self->menu, self);
       g_clear_object (&self->menu);
     }
@@ -219,15 +232,238 @@ indicator_ng_secondary_activate (IndicatorObject      *io,
     }
 }
 
-static void
-indicator_ng_menu_shown (GtkWidget *widget,
-                         gpointer   user_data)
+static gboolean indicator_ng_menu_insert_idos(IndicatorNg *self, GMenuModel *pSection, guint nModelItem, guint nMenuItem, gboolean bNamespace, gchar *sNamespace)
 {
-  IndicatorNg *self = user_data;
+    gboolean bChanged = FALSE;
+    gchar *sType;
+    gboolean bHasType = g_menu_model_get_item_attribute(pSection, nModelItem, "x-ayatana-type", "s", &sType);
 
-  if (self->submenu_action)
-    g_action_group_change_action_state (self->actions, self->submenu_action,
-                                        g_variant_new_boolean (TRUE));
+    if (bHasType)
+    {
+        GList *lMenuItems = gtk_container_get_children(GTK_CONTAINER(self->entry.menu));
+        GtkWidget *pMenuItemOld = GTK_WIDGET(g_list_nth_data(lMenuItems, nMenuItem));
+        const gchar *sName = gtk_widget_get_name(pMenuItemOld);
+
+        if (!g_str_equal(sName, sType))
+        {
+            GActionGroup *pActionGroup = (GActionGroup*)g_object_get_qdata(G_OBJECT(self->entry.menu), m_pActionMuxer);
+            GMenuItem *pMenuModelItem = g_menu_item_new_from_model(pSection, nModelItem);
+            GtkMenuItem* pMenuItemNew = NULL;
+            gchar *sAction;
+
+            if (bNamespace && g_menu_item_get_attribute(pMenuModelItem, G_MENU_ATTRIBUTE_ACTION, "s", &sAction))
+            {
+                gchar *sNamespacedAction = g_strconcat(sNamespace, ".", sAction, NULL);
+                g_menu_item_set_attribute(pMenuModelItem, G_MENU_ATTRIBUTE_ACTION, "s", sNamespacedAction);
+                g_free (sNamespacedAction);
+                g_free (sAction);
+            }
+
+            for (GList *pFactory = ayatana_menu_item_factory_get_all(); pFactory != NULL && pMenuItemNew == NULL; pFactory = pFactory->next)
+            {
+                pMenuItemNew = ayatana_menu_item_factory_create_menu_item(pFactory->data, sType, pMenuModelItem, pActionGroup);
+                bChanged = TRUE;
+            }
+
+            if (pMenuItemNew == NULL)
+            {
+                pMenuItemNew = GTK_MENU_ITEM(gtk_menu_item_new_with_label("Failed to create IDO object"));
+            }
+
+            gtk_widget_set_name(GTK_WIDGET(pMenuItemNew), sType);
+            gtk_widget_show(GTK_WIDGET(pMenuItemNew));
+            gtk_container_remove(GTK_CONTAINER(self->entry.menu), pMenuItemOld);
+            gtk_menu_shell_insert(GTK_MENU_SHELL(self->entry.menu), GTK_WIDGET(pMenuItemNew), nMenuItem);
+            g_object_unref(pMenuModelItem);
+        }
+
+        g_list_free(lMenuItems);
+        g_free(sType);
+    }
+
+    return bChanged;
+}
+
+static void indicator_ng_menu_size_allocate(GtkWidget *pWidget, GtkAllocation *pAllocation, gpointer pUserData)
+{
+    IndicatorNg *self = pUserData;
+    GList *pMenuItem = gtk_container_get_children(GTK_CONTAINER(self->entry.menu));
+    guint nWidth = 0;
+    guint nHeight = 0;
+    GdkWindow *pWindowBin = NULL;
+
+    while (pMenuItem)
+    {
+        if (!pWindowBin)
+        {
+            pWindowBin = gtk_widget_get_parent_window(pMenuItem->data);
+        }
+
+        gint nWidthNat;
+        gint nHeightNat;
+        gtk_widget_get_preferred_width(pMenuItem->data, NULL, &nWidthNat);
+        gtk_widget_get_preferred_height(pMenuItem->data, NULL, &nHeightNat);
+        nWidth = MAX(nWidth, nWidthNat);
+        nHeight += nHeightNat;
+        GtkBorder cPadding;
+        GtkStyleContext *pContext = gtk_widget_get_style_context(GTK_WIDGET(pMenuItem->data));
+        gtk_style_context_get_padding(pContext, gtk_style_context_get_state(pContext), &cPadding);
+        nWidth += cPadding.left + cPadding.right;
+        pMenuItem = g_list_next(pMenuItem);
+    }
+
+    g_list_free(pMenuItem);
+    GtkBorder cPadding;
+    GtkStyleContext *pContext = gtk_widget_get_style_context(GTK_WIDGET(self->entry.menu));
+    gtk_style_context_get_padding(pContext, gtk_style_context_get_state(pContext), &cPadding);
+    gint nBorderWidth = gtk_container_get_border_width(GTK_CONTAINER(self->entry.menu));
+    gint nIconWidth;
+    gtk_icon_size_lookup(GTK_ICON_SIZE_MENU, &nIconWidth, NULL);
+    nWidth += (2 * nBorderWidth) + cPadding.left + cPadding.right + (nIconWidth * 3) / 2;
+    nHeight += (2 * nBorderWidth) + cPadding.top + cPadding.bottom + (nIconWidth * 3) / 4;
+
+    GdkRectangle cRectangle = {0};
+    GdkDisplay *pDisplay = gdk_display_get_default();
+    GdkMonitor *pMonitor = gdk_display_get_primary_monitor(pDisplay);
+    gdk_monitor_get_workarea(pMonitor, &cRectangle);
+
+    if (nHeight <= cRectangle.height)
+    {
+        gdk_window_move_resize(pWindowBin, 0, 0, nWidth, nHeight);
+    }
+
+    nHeight = MIN(nHeight, cRectangle.height);
+
+    GdkWindow *pWindow = gtk_widget_get_parent_window(GTK_WIDGET(self->entry.menu));
+    gdk_window_resize(pWindow, nWidth, nHeight);
+
+    gtk_menu_reposition(self->entry.menu);
+}
+
+static void indicator_ng_menu_section_changed(GMenuModel *pMenuSection, gint nPosition, gint nRemoved, gint nAdded, gpointer pUserData)
+{
+    IndicatorNg *self = pUserData;
+    GMenuModel *pModel = g_menu_model_get_item_link(self->menu, 0, G_MENU_LINK_SUBMENU);
+    guint nMenuItem = 0;
+    gboolean bChanged = FALSE;
+
+    if (pModel)
+    {
+        guint nSections = g_menu_model_get_n_items(pModel);
+
+        for (guint nSection = 0; nSection < nSections; nSection++)
+        {
+            GMenuModel *pSection = g_menu_model_get_item_link(pModel, nSection, G_MENU_LINK_SECTION);
+
+            if (pSection)
+            {
+                gchar *sNamespace;
+                gboolean bNamespace = g_menu_model_get_item_attribute(pModel, nSection, G_MENU_ATTRIBUTE_ACTION_NAMESPACE, "s", &sNamespace);
+                guint nSubsections = g_menu_model_get_n_items(pSection);
+
+                for (guint nSubsection = 0; nSubsection < nSubsections; nSubsection++)
+                {
+                    GMenuModel *pSubsection = g_menu_model_get_item_link(pSection, nSubsection, G_MENU_LINK_SECTION);
+
+                    if (pSubsection)
+                    {
+                        guint nItems = g_menu_model_get_n_items(pSubsection);
+
+                        // Skip the subsection separator (if there is one)
+                        GList *lMenuItems = gtk_container_get_children(GTK_CONTAINER(self->entry.menu));
+                        GtkWidget *pMenuItem = GTK_WIDGET(g_list_nth_data(lMenuItems, nMenuItem));
+
+                        if (GTK_IS_SEPARATOR_MENU_ITEM(pMenuItem))
+                        {
+                            nMenuItem++;
+                        }
+
+                        g_list_free(lMenuItems);
+
+                        for (guint nItem = 0; nItem < nItems; nItem++)
+                        {
+                            bChanged = indicator_ng_menu_insert_idos(self, pSubsection, nItem, nMenuItem, bNamespace, sNamespace) || bChanged;
+                            nMenuItem++;
+                        }
+                    }
+
+                    g_object_unref(pSubsection);
+
+                    bChanged = indicator_ng_menu_insert_idos(self, pSection, nSubsection, nMenuItem, bNamespace, sNamespace) || bChanged;
+
+                    if (!g_str_equal(self->name, "ayatana-indicator-messages"))
+                    {
+                        nMenuItem++;
+                    }
+                }
+
+                if (bNamespace)
+                {
+                    g_free(sNamespace);
+                }
+
+                g_object_unref(pSection);
+            }
+
+            if (!g_str_equal(self->name, "ayatana-indicator-messages"))
+            {
+                nMenuItem++;
+            }
+        }
+
+        g_object_unref(pModel);
+    }
+
+    if (bChanged)
+    {
+        indicator_ng_menu_size_allocate(NULL, NULL, self);
+    }
+}
+
+static void indicator_ng_menu_shown(GtkWidget *pWidget, gpointer pUserData)
+{
+    IndicatorNg *self = pUserData;
+    guint nSectionCount = 0;
+
+    if (!self->lMenuSections[0])
+    {
+        self->lMenuSections[0] = g_menu_model_get_item_link(self->menu, 0, G_MENU_LINK_SUBMENU);
+
+        if (self->lMenuSections[0])
+        {
+            guint nSections = g_menu_model_get_n_items(self->lMenuSections[0]);
+
+            for (gint nSection = 0; nSection < nSections; nSection++)
+            {
+                self->lMenuSections[++nSectionCount] = g_menu_model_get_item_link(self->lMenuSections[0], nSection, G_MENU_LINK_SECTION);
+
+                if (self->lMenuSections[nSectionCount])
+                {
+                    g_signal_connect(self->lMenuSections[nSectionCount], "items-changed", G_CALLBACK(indicator_ng_menu_section_changed), self);
+                    guint nSubsections = g_menu_model_get_n_items(self->lMenuSections[nSectionCount]);
+                    guint nParent = nSectionCount;
+
+                    for (guint nSubsection = 0; nSubsection < nSubsections; nSubsection++)
+                    {
+                        self->lMenuSections[++nSectionCount] = g_menu_model_get_item_link(self->lMenuSections[nParent], nSubsection, G_MENU_LINK_SECTION);
+
+                        if (self->lMenuSections[nSectionCount])
+                        {
+                            g_signal_connect(self->lMenuSections[nSectionCount], "items-changed", G_CALLBACK(indicator_ng_menu_section_changed), self);
+                        }
+                    }
+                }
+            }
+
+            g_signal_connect(self->lMenuSections[0], "items-changed", G_CALLBACK(indicator_ng_menu_section_changed), self);
+            indicator_ng_menu_section_changed(self->lMenuSections[0], 0, 0, 1, self);
+        }
+    }
+
+    if (self->submenu_action)
+    {
+        g_action_group_change_action_state(self->actions, self->submenu_action, g_variant_new_boolean(TRUE));
+    }
 }
 
 static void
@@ -358,7 +594,7 @@ indicator_ng_menu_item_is_of_type (GMenuModel  *menu,
   gchar *type;
   gboolean has_type = FALSE;
 
-  if (g_menu_model_get_item_attribute (menu, index, "x-canonical-type", "s", &type))
+  if (g_menu_model_get_item_attribute (menu, index, "x-ayatana-type", "s", &type))
     {
       has_type = g_str_equal (type, expected_type);
       g_free (type);
@@ -404,14 +640,14 @@ indicator_ng_menu_changed (GMenuModel *menu,
               g_free (action);
             }
 
-          if (g_menu_model_get_item_attribute (self->menu, 0, "x-canonical-scroll-action", "s", &action))
+          if (g_menu_model_get_item_attribute (self->menu, 0, "x-ayatana-scroll-action", "s", &action))
             {
               if (g_str_has_prefix (action, "indicator."))
                 self->scroll_action = g_strdup (action + strlen ("indicator."));
               g_free (action);
             }
 
-          if (g_menu_model_get_item_attribute (self->menu, 0, "x-canonical-secondary-action", "s", &action))
+          if (g_menu_model_get_item_attribute (self->menu, 0, "x-ayatana-secondary-action", "s", &action))
             {
               if (g_str_has_prefix (action, "indicator."))
                 self->secondary_action = g_strdup (action + strlen ("indicator."));
@@ -425,6 +661,14 @@ indicator_ng_menu_changed (GMenuModel *menu,
               g_free (action);
             }
 
+            for (guint nMenuSection = 0; nMenuSection < MENU_SECTIONS; nMenuSection++)
+            {
+                if (self->lMenuSections[nMenuSection])
+                {
+                    g_object_unref(self->lMenuSections[nMenuSection]);
+                }
+            }
+
           popup = g_menu_model_get_item_link (self->menu, 0, G_MENU_LINK_SUBMENU);
           if (popup)
             {
@@ -687,6 +931,13 @@ indicator_ng_initable_iface_init (GInitableIface *initable)
 static void
 indicator_ng_init (IndicatorNg *self)
 {
+    m_pActionMuxer = g_quark_from_static_string ("gtk-widget-action-muxer");
+
+    for (guint nMenuSection = 0; nMenuSection < MENU_SECTIONS; nMenuSection++)
+    {
+        self->lMenuSections[nMenuSection] = NULL;
+    }
+
   self->entry.label = (GtkLabel*)g_object_ref_sink (gtk_label_new (NULL));
   self->entry.image = (GtkImage*)g_object_ref_sink (gtk_image_new ());
 
@@ -694,6 +945,7 @@ indicator_ng_init (IndicatorNg *self)
 
   g_signal_connect (self->entry.menu, "show", G_CALLBACK (indicator_ng_menu_shown), self);
   g_signal_connect (self->entry.menu, "hide", G_CALLBACK (indicator_ng_menu_hidden), self);
+  g_signal_connect (self->entry.menu, "size-allocate", G_CALLBACK (indicator_ng_menu_size_allocate), self);
 
   /* work around IndicatorObject's warning that the accessible
    * description is missing. We never set it on construction, but when
